"use strict";
const C3 = self.C3
const ROT = globalThis.ROT
{
	C3.Plugins.aaXe_RotJs.Acts = {
		SetSeed(seed) {
			//set the seed => enable the ssed
			this._useSeed = true;
			this._seed = seed;
			ROT.RNG.setSeed(this._seed);
		},

		DebugRot() {
			console.log('seed =>');
			console.log(this._seed);
			console.log('map =>');
			console.log(this.map);
			console.log('rooms =>');
			console.log(this.rooms);
			console.log('doors =>');
			console.log(this.doors);
		},

		DebugFov() {
			console.log("fov => ");
			console.log(this.fovMap);
		},

		GenerateArenaToTileMap(width, height, tm) {
			this._clearRooms();
			this.width = width;
			this.height = height;
			var instance = tm.GetInstances()[0].GetSdkInstance();
			this.gen = new ROT.Map.Arena(width, height);
			this.map = {};

			this.gen.create((x, y, value) => {
				var key = x + "," + y;
				this.map[key] = {
					x: x,
					y: y,
					value: value,
					type: value === 1 ? 'wall' : 'floor'
				};

				//populate tilemap
				value ? instance.SetTileAt(x, y, this._getWallTile()) : instance.SetTileAt(x, y, this._getFloorTile());			
			});

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.MapGenerated);
			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.ArenaGenerated);
		},

		GenerateDividedMazeToTileMap(width, height, tm) {
			this._clearRooms();
			this.width = width;
			this.height = height;
			var instance = tm.GetInstances()[0].GetSdkInstance();
			this.gen = new ROT.Map.DividedMaze(width, height);
			this.map = {};

			this.gen.create((x, y, value) => {
				var key = x + "," + y;
				this.map[key] = {
					x: x,
					y: y,
					value: value,
					type: value === 1 ? 'wall' : 'floor'
				};

				//populate tilemap
				value ? instance.SetTileAt(x, y, this._getWallTile()) : instance.SetTileAt(x, y, this._getFloorTile());			
			});

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.MapGenerated);
			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.MazeGenerated);
		},

		GenerateIceyMazeToTileMap(width, height, reg, tm) {
			this._clearRooms();
			this.width = width;
			this.height = height;
			var instance = tm.GetInstances()[0].GetSdkInstance();
			this.gen = new ROT.Map.IceyMaze(width, height, reg);
			this.map = {};

			this.gen.create((x, y, value) => {
				var key = x + "," + y;
				this.map[key] = {
					x: x,
					y: y,
					value: value,
					type: value === 1 ? 'wall' : 'floor'
				};

				//populate tilemap
				value ? instance.SetTileAt(x, y, this._getWallTile()) : instance.SetTileAt(x, y, this._getFloorTile());			
			});

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.MapGenerated);
			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.MazeGenerated);
		},

		GenerateEllersPerfectMazeToTileMap(width, height, tm) {
			this._clearRooms();
			this.width = width;
			this.height = height;
			var instance = tm.GetInstances()[0].GetSdkInstance();
			this.gen = new ROT.Map.EllerMaze(width, height);
			this.map = {};

			this.gen.create((x, y, value) => {
				var key = x + "," + y;
				this.map[key] = {
					x: x,
					y: y,
					value: value,
					type: value === 1 ? 'wall' : 'floor'
				};

				//populate tilemap
				value ? instance.SetTileAt(x, y, this._getWallTile()) : instance.SetTileAt(x, y, this._getFloorTile());			
			});

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.MapGenerated);
			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.MazeGenerated);
		},

		GenerateCellularAutomataToTileMap(width, height, rnd, it, conn, tm) {
			this._clearRooms();
			this.width = width;
			this.height = height;
			var instance = tm.GetInstances()[0].GetSdkInstance();
			this.gen = new ROT.Map.Cellular(width, height, {
				connected: conn === 0 ? true : false
			});
			this.gen.randomize(rnd);
			this.map = {};

			//iterate
			for (var i = 0; i < it; i++) {
				//last iteration, draw map
				if (i == it - 1) {
					this.gen.create((x, y, value) => {
						var key = x + "," + y;
						this.map[key] = {
							x: x,
							y: y,
							value: value,
							type: value === 1 ? 'wall' : 'floor'
						};
						value ? instance.SetTileAt(x, y, this._getWallTile()) : instance.SetTileAt(x, y, this._getFloorTile());			
					});
				} else {
					this.gen.create((x, y, value) => {
						var key = x + "," + y;
						this.map[key] = {
							x: x,
							y: y,
							value: value,
							type: value === 1 ? 'wall' : 'floor'
						};
					});
				}
			}

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.MapGenerated);
			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.CellularAutomataGenerated);
		},

		GenerateCellularAutomataAdvancedToTileMap(width, height, rnd, born, survive, it, conn, tm) {
			this._clearRooms();
			this.width = width;
			this.height = height;
			var instance = tm.GetInstances()[0].GetSdkInstance();
			var b = born.split(",").map(x => Number(x));
			var s = survive.split(",").map(x => Number(x));

			this.gen = new ROT.Map.Cellular(width, height, {
				connected: conn === 0 ? true : false,
				born: b,
				survive: s
			});
			this.gen.randomize(rnd);
			this.map = {};

			//iterate
			for (var i = 0; i < it; i++) {
				//last iteration, draw map
				if (i == it - 1) {
					this.gen.create((x, y, value) => {
						var key = x + "," + y;
						this.map[key] = {
							x: x,
							y: y,
							value: value,
							type: value === 1 ? 'wall' : 'floor'
						};
						value ? instance.SetTileAt(x, y, this._getWallTile()) : instance.SetTileAt(x, y, this._getFloorTile());			
					});
				} else {
					this.gen.create((x, y, value) => {
						var key = x + "," + y;
						this.map[key] = {
							x: x,
							y: y,
							value: value,
							type: value === 1 ? 'wall' : 'floor'
						};

					});
				}
			}

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.MapGenerated);
			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.CellularAutomataGenerated);
		},

		GenerateDiggerToTileMap(width, height, tm) {
			this._clearRooms();
			this.width = width;
			this.height = height;
			var instance = tm.GetInstances()[0].GetSdkInstance();

			this.gen = new ROT.Map.Digger(width, height);
			this.map = {};

			this.gen.create((x, y, value) => {

				var key = x + "," + y;
				this.map[key] = {
					x: x,
					y: y,
					value: value,
					type: value === 1 ? 'wall' : 'coor'
				};
				
				value ? instance.SetTileAt(x, y, this._getWallTile()) : instance.SetTileAt(x, y, this._getFloorTile());			
			});

			//get rooms info
			var r = this.gen.getRooms();
			for (var i = 0; i < r.length; i++) {
				var room = r[i];
				this._setupRooms(room, i);
			}

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.MapGenerated);
			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.DungeonGenerated);
		},

		GenerateDiggerAdvancedToTileMap(width, height, minrw, maxrw, minrh, maxrh, mincl, maxcl, dp, to, tm) {
			this._clearRooms();
			this.width = width;
			this.height = height;
			var instance = tm.GetInstances()[0].GetSdkInstance();

			this.gen = new ROT.Map.Digger(width, height, {
				roomWidth: [minrw, maxrw],
				roomHeight: [minrh, maxrh],
				corridorLength: [mincl, maxcl],
				dugPercentage: dp,
				timeLimit: to
			});
			this.map = {};

			this.gen.create((x, y, value) => {
				var key = x + "," + y;
				this.map[key] = {
					x: x,
					y: y,
					value: value,
					type: value === 1 ? 'wall' : 'coor'
				};

				value ? instance.SetTileAt(x, y, this._getWallTile()) : instance.SetTileAt(x, y, this._getFloorTile());			
			});

			//get rooms info
			var r = this.gen.getRooms();
			for (var i = 0; i < r.length; i++) {
				var room = r[i];
				this._setupRooms(room, i);
			}

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.MapGenerated);
			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.DungeonGenerated);
		},

		GenerateUniformToTileMap(width, height, tm) {
			this._clearRooms();
			this.width = width;
			this.height = height;
			var instance = tm.GetInstances()[0].GetSdkInstance();

			this.gen = new ROT.Map.Uniform(width, height);
			this.map = {};

			this.gen.create((x, y, value) => {
				var key = x + "," + y;
				this.map[key] = {
					x: x,
					y: y,
					value: value,
					type: value === 1 ? 'wall' : 'coor'
				};

				value ? instance.SetTileAt(x, y, this._getWallTile()) : instance.SetTileAt(x, y, this._getFloorTile());			
			});

			//get rooms info
			var r = this.gen.getRooms();
			for (var i = 0; i < r.length; i++) {
				var room = r[i];
				this._setupRooms(room, i);
			}

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.MapGenerated);
			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.DungeonGenerated);
		},

		GenerateUniformAdvancedToTileMap(width, height, minrw, maxrw, minrh, maxrh, dp, to, tm) {
			this._clearRooms();
			this.width = width;
			this.height = height;
			var instance = tm.GetInstances()[0].GetSdkInstance();

			this.gen = new ROT.Map.Uniform(width, height, {
				roomWidth: [minrw, maxrw],
				roomHeight: [minrh, maxrh],
				dugPercentage: dp,
				timeLimit: to
			});
			this.map = {};

			this.gen.create((x, y, value) => {
				var key = x + "," + y;
				this.map[key] = {
					x: x,
					y: y,
					value: value,
					type: value === 1 ? 'wall' : 'coor'
				};

				value ? instance.SetTileAt(x, y, this._getWallTile()) : instance.SetTileAt(x, y, this._getFloorTile());			
			});

			//get rooms info
			var r = this.gen.getRooms();
			for (var i = 0; i < r.length; i++) {
				var room = r[i];
				this._setupRooms(room, i);
			}

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.MapGenerated);
			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.DungeonGenerated);
		},

		GenerateRogueToTileMap(width, height, tm) {
			this._clearRooms();
			this.width = width;
			this.height = height;
			var instance = tm.GetInstances()[0].GetSdkInstance();

			this.gen = new ROT.Map.Rogue(width, height);
			this.map = {};

			this.gen.create((x, y, value) => {
				var key = x + "," + y;
				this.map[key] = {
					x: x,
					y: y,
					value: value,
					type: value === 1 ? 'wall' : 'floor'
				};

				value ? instance.SetTileAt(x, y, this._getWallTile()) : instance.SetTileAt(x, y, this._getFloorTile());			
			});

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.MapGenerated);
			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.DungeonGenerated);
		},

		DownloadMapAsJson() {
			var data = this.SaveToJson();
			var json = JSON.stringify(data);
			var blob = new Blob([json], {
				type: "application/json"
			});
			var url = URL.createObjectURL(blob);

			const runtime = this._runtime;
			runtime.InvokeDownload(url, 'map.json');
		},

		LoadJsonToTilemap(json, tm) {
			var o = JSON.parse(json);
			this.LoadFromJson(o);
			var instance = tm.GetInstances()[0].GetSdkInstance();

			var mapValues = Object.values(this.map);
			for (const cell of mapValues) {
				instance.SetTileAt(cell.x, cell.y, cell.value);
			}
		},

		LoadMapFromJson(json) {
			var o = JSON.parse(json);
			this.LoadFromJson(o);
		},

		GenerateBlobIndex(cellValue) {
			//create new visited 2d array [width][height]
			var matrix = new Array(this.width).fill(-1).map(() => new Array(this.height).fill(-1));
			var visited = new Array(this.width).fill(-1).map(() => new Array(this.height).fill(-1));

			var count = 0;
			for (var x = 0; x < this.width; x++) {
				for (var y = 0; y < this.height; y++) {
					//check cell is not vistsed
					if (matrix[x][y] == -1 && visited[x][y] == -1 && this.map[x + ',' + y].value === cellValue) {
						this._visitCell(x, y, count, matrix, visited, cellValue);
						count++;
					}
				}
			}

			//assign blob maps
			for (var x = 0; x < this.width; x++) {
				for (var y = 0; y < this.height; y++) {
					var index = matrix[x][y];
					if (index >= 0) {

						if (this.blobs[index]) {
							this.blobs[index].push({
								x: x,
								y: y
							});
						} else {
							this.blobs[index] = [];
							this.blobs[index].push({
								x: x,
								y: y
							});
						}
					}
				}
			}

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.BlobsGenerated);
		},

		GenerateFovPerciseShadowcasting(x, y, rad) {
			this.fovMap = [];

			var fov = new ROT.FOV.PreciseShadowcasting((x, y) => {
				if (x + ',' + y in this.map) {
					//if value is 0 (empty space) light is allowed to pass
					return (this.map[x + ',' + y].value === 0);
				}
				return false;
			});

			fov.compute(x, y, rad, (x, y, r, v) => {
				this.fovMap.push({
					x: x,
					y: y,
					radius: r,
					visibility: v
				});
			});

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.FovGenerated);
		},

		GenerateFovRecursiveShadowcasting360(x, y, rad) {
			this.fovMap = [];

			var fov = new ROT.FOV.RecursiveShadowcasting((x, y) => {
				if (x + ',' + y in this.map) {
					//if value is 0 (empty space) light is allowed to pass
					return (this.map[x + ',' + y].value === 0);
				}
				return false;
			});

			fov.compute(x, y, rad, (x, y, r, v) => {
				this.fovMap.push({
					x: x,
					y: y,
					radius: r,
					visibility: v
				});
			});

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.FovGenerated);
		},

		GenerateFovRecursiveShadowcasting180(x, y, rad, dir) {		
			this.fovMap = [];

			//direction 0=north, 1=northeast, 2=east, 3=southeast, 4=south, 5=southwest, 6=west, 7=northwest
			var dirEnum = {
				N:0, NE:1, E:2, SE:3, S:4, SW:5, W:6, NW:7
			};

			var fov = new ROT.FOV.RecursiveShadowcasting((x, y) => {
				if (x + ',' + y in this.map) {
					//if value is 0 (empty space) light is allowed to pass
					return (this.map[x + ',' + y].value === 0);
				}
				return false;
			});

			let direction = dirEnum[dir.toUpperCase()];
			console.log(direction);
			if (direction === null) {
				console.log('error => invalid direction code');
				return;
			}

			fov.compute180(x, y, rad, direction, (x, y, r, v) => {
				this.fovMap.push({
					x: x,
					y: y,
					radius: r,
					visibility: v
				});
			});

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.FovGenerated);
		},

		GenerateFovRecursiveShadowcasting90(x, y, rad, dir) {
			this.fovMap = [];

			//direction 0=north, 1=northeast, 2=east, 3=southeast, 4=south, 5=southwest, 6=west, 7=northwest
			var dirEnum = {
				N:0, NE:1, E:2, SE:3, S:4, SW:5, W:6, NW:7
			};

			var fov = new ROT.FOV.RecursiveShadowcasting((x, y) => {
				if (x + ',' + y in this.map) {
					//if value is 0 (empty space) light is allowed to pass
					return (this.map[x + ',' + y].value === 0);
				}
				return false;
			});

			let direction = dirEnum[dir.toUpperCase()];
			console.log(direction);
			if (direction === null) {
				console.log('error => invalid direction code');
				return;
			}

			fov.compute90(x, y, rad, direction, (x, y, r, v) => {
				this.fovMap.push({
					x: x,
					y: y,
					radius: r,
					visibility: v
				});
			});

			this.Trigger(C3.Plugins.aaXe_RotJs.Cnds.FovGenerated);
		},

		AddFloorTile(index, weight)
		{
			if(this.floorCleared){
				this.floorTiles = {};
				this.floorCleared = false;
			}
			//todo: fix this use objetc property
			this.floorTiles[index] = weight;
		},

		AddWallTile(index, weight)
		{
			if(this.wallCleared){
				this.wallTiles = {};
				this.wallCleared = false;
			}
			//todo: fix this use objetc property
			this.wallTiles[index] = weight;
		},

		ClearFloorTile()
		{
			this.floorCleared = true;
			this.floorTiles = { "0" : "100"};
		},

		ClearWallTile()
		{
			this.wallCleared = true;
			this.wallTiles =  { "1" : "100"};
		}
	};
}

//todo: add action to set default wall tiles and default floor tile index